/**
 * ContextSensitive: a library for generating context-sensitive
 * content on HTML elements. It will take over the
 * onmousedown/oncontextmenu functions for the document, and works
 * only where these are possible to override.
 *
 * The classes rely on prototype.js (http://www.prototypejs.org/) and
 * are released under the same terms as the prototype library.
 *
 * Original code by Havard Eide (http://eide.org/) released under the MIT
 * license.
 *
 * $Horde: dimp/js/src/ContextSensitive.js,v 1.19 2007/04/18 15:17:39 slusarz Exp $
 *
 * @author Chuck Hagenbuch <chuck@horde.org>
 */

var ContextSensitive = Class.create();
ContextSensitive.prototype = {
    // Vars used and defaulting to null/false:
    //   current, target
    elements: [],

    initialize: function()
    {
        Event.observe(document, 'contextmenu', this.rightclick.bindAsEventListener(this));
        Event.observe(document, 'mousedown', this.mouseclick.bindAsEventListener(this));
    },

    /**
     * Elements are of type ContextSensitive.Element.
     */
    addElement: function(id, target, offset)
    {
        if (DMenu.validElement(id)) {
            return this.valid(id);
        }
        var el = new ContextSensitive.Element(id, target, offset);
        this.elements.push(el);
        return el;
    },

    /**
     * Remove a registered element.
     */
    removeElement: function(id)
    {
        this.elements = this.elements.findAll(function(s) {
            return s.id != id;
        });
    },

    /**
     * Hide the current element. Call this when working on an element
     * and you want to close it after some work has been done:
     *
     * <a onclick="highlightText('javascript'); ctx.close();">Highlight</a>
     */
    close: function()
    {
        if (this.current) {
            this.current.hide();
        }
    },

    /**
     * Get the element that was clicked on to trigger the current
     * context menu (if any).
     */
    element: function()
    {
        return this.target;
    },

    /**
     * Get a valid element (the ones that can be right-clicked) based
     * on a element ID.
     */
    validElement: function(id)
    {
        return this.elements.detect(function(s) {
            return s.id == id;
        });
    },

    /**
     * Get a element based on the element ID (the one to display).
     */
    valid: function(id)
    {
        return this.elements.detect(function(s) {
            return s.ctx == id;
        });
    },

    /**
     * Called when a click is registered. Will return before the
     * element is closed if we click on an element inside of it.
     */
    mouseclick: function(e)
    {
        if (this.current && this.current.visible()) {
            var c_id = this.current.id, target = Event.element(e);
            for (; target; target = target.parentNode) {
                if (target.id && target.id == c_id) {
                    return;
                }
            }
            this.close();
        }
        return true;
    },

    /**
     * Called when a rightclick is registered.
     */
    rightclick: function(e)
    {
        // Make sure we hide any other ContextObjects.
        this.mouseclick(e);

        // What html element have we received a rightclick on?
        var target;
        if (e.target) {
            target = e.target;
            if (target.nodeType == 3) {
                target = target.parentNode;
            }
        } else if (e.srcElement) {
            target = e.srcElement;
        }

        // Get the rightclick element based on the ID of the element
        // we just have clicked.
        var ctx = this.validElement(target.id);
        if (!ctx) {
            while (target.parentNode && !ctx) {
                target = target.parentNode;
                ctx = this.validElement(target.id);
            }
            if (!ctx) {
                return;
            }
        }

        // Register the current element that will be shown and the
        // element that was clicked on.
        this.current = $(ctx.ctx);
        this.target = target;

        // Try to retrieve the context-sensitive element we want to
        // display. If we can't find it we just return.
        var el = $(ctx.ctx);
        if (!el) {
            return;
        }

        // Get the base element positions.
        var l, left, pos, size, t;
        if (ctx.offset) {
            var offsets = Position.positionedOffset(target);
            l = offsets[0];
            t = offsets[1] + target.getHeight();
        } else {
            l = Event.pointerX(e);
            t = Event.pointerY(e);
        }

        // Get window dimensions
        var de = document.documentElement;
        var w = window.innerWidth || self.innerWidth || (de && de.clientWidth) || document.body.clientWidth,
            h = window.innerHeight || self.innerHeight || (de && de.clientHeight) || document.body.clientHeight;

        // We need to make the context element active in order to determine
        // its height and width.  However, position the element well off
        // screen so the user doesn't see it.
        pos = el.getStyle('position'), 
        el.setStyle({ left: '-999em', position: 'absolute', display: '' });
        size = el.getDimensions();

        // Make sure context window is entirely on screen
        if ((t + size.height) > h) {
            t = h - size.height - 10;
        }
        if ((l + size.width) > w) {
            l = w - size.width - 10;
        }

        el.setStyle({ display: 'block', left: l + 'px', top: t + 'px', position: pos });

        // Stop event to prevent browser context menu from displaying
        Event.stop(e);
    }
};

ContextSensitive.Element = Class.create();
ContextSensitive.Element.prototype = {
    initialize: function(id, target, offset)
    {
        this.id = id;
        this.ctx = target;
        this.offset = offset;

        var t = $(target);
        t.hide();
        t.setStyle({ position: 'absolute' });
    }
};
