/**
 * ViewPort.js - Code to create a viewport window in a web browser.
 *
 * Requires prototypejs 1.5.1+ and scriptaculous 1.5+
 *
 * $Horde: dimp/js/src/ViewPort.js,v 1.92 2007/04/20 17:40:13 slusarz Exp $
 *
 * Copyright 2005-2007 The Horde Project (http://www.horde.org/)
 *
 * See the enclosed file COPYING for license information (GPL). If you
 * did not receive this file, see http://www.fsf.org/copyleft/gpl.html.
 */

/**
 * ViewPort
 */
ViewPort = Class.create();
ViewPort.prototype = {
    lastDisplayedStartPos: -1,
    lastDisplayedPageSize: -1,
    views: {},

    initialize: function(container, lines, totalRows, action, options) {
        this.containers = {
            content: container.content ? $(container.content) : $(container),
            empty: container.empty ? $(container.empty) : null,
            error: container.error ? $(container.error) : null
        };
        this.lines = lines;
        this.totalRows = totalRows;
        this.action = action;
        this.options = options || {};
    },

    loadView: function(id, params, uid) {
        this.clearWait();

        if (this.metaData) {
            this.views[this.id] = { metaData: this.metaData, buffer: this.buffer, offset: this.currentOffset() };
        }

        this.clear();
        this.additionalParams = params;
        this.id = id;

        if (this.views[id]) {
            this.metaData = this.views[id].metaData;
            this.buffer = this.views[id].buffer;
            this.checkCache();
            if (this.filter && this.filter.filtering) {
                this.filter.filter();
            } else {
                this.updateContent(this.views[id].offset);
            }
            if (this.options.onComplete) {
                this.options.onComplete();
            }
            return true;
        }

        this.metaData = new ViewPort.MetaData(this.lines, this.totalRows);
        this.buffer = new ViewPort.Buffer(this);
        if (uid) {
            this.fetchBuffer(null, { uid: uid });
        } else {
            this.fetchBuffer(0);
        }
        this.clearWait();
        return false;
    },

    getRowById: function(id) {
        return (this.buffer) ? this.buffer.getRowById(id) : null;
    },

    positionById: function(id) {
        var row = this.getRowById(id);
        if (!row) {
            return false;
        }

        var num = row.rownum;
        var smaller = num < this.lastDisplayedStartPos;
        var larger = num >= this.lastDisplayedStartPos + this.metaData.get('curr_page_size');
 
        if ((smaller || larger) && this.scroller) {
            num = (smaller)
                ? num
                : num - this.metaData.get('page_size') + 1;
            this.scroller.moveScroll(num);
        }

        return true;
    },

    reload: function(params) {
        this.fetchBuffer(this.scroller ? this.scroller.currentOffset() : 0, params);
    },

    addFilter: function(input, action, callback) {
        this.filter = new ViewPort.Filter(this, input, action, callback);
    },

    runFilter: function() {
        if (this.filter) {
            this.filter.filter(true);
        }
    },

    setTotalRows: function(n) {
        this.metaData.set({ total_rows: n });
        if (this.scroller) {
            this.scroller.updateSize();
        }
    },

    onResize: function() {
        if (this.scroller) {
            this.scroller.onResize();
        }
    },

    largeBufferWindowStart: function(offset) {
        var val = offset - ((0.5 * this.metaData.get('buffer_size')) - (0.5 * this.metaData.get('page_size')));
        return Math.max(parseInt(val), 0);
    },

    requestContentRefresh: function(offset, opts) {
        opts = opts || {};
        if (offset == this.lastDisplayedStartPos &&
            this.metaData.get('page_size') == this.lastDisplayedPageSize) {
            return false;
        }
        if (this.buffer.isLoaded(offset)) {
            this.updateContent(offset);
            if (this.buffer.isNearingLimit(offset)) {
                this.fetchBuffer(offset, Object.extend(opts, { noUpdateContent: true }));
            }
            return true;
        } else {
            this.fetchBuffer(offset, opts);
            return false;
        }
    },

    fetchBuffer: function(offset, fetchOptions) {
        fetchOptions = fetchOptions || {};

        if (this.options.onFetch) {
            this.options.onFetch();
        }

        var callParams = $H({ page_size: this.metaData.get('buffer_size') }).merge(this.additionalParams).merge(fetchOptions);
        // If uid exists, don't use offset
        if (!callParams.uid) {
            callParams.offset = this.largeBufferWindowStart(offset);
        }

        DimpCore.doAction(this.action, callParams, [], function(r) { this.ajaxUpdate(r, { offset: offset, options: fetchOptions, id: this.id }) }.bind(this), Object.clone(this.options));
        this.handleWait();
    },

    ajaxUpdate: function(r, opts) {
        this.clearWait();

        if (!r.response) {
            if (this.id == opts.id) {
                if (this.options.onFail) {
                    this.options.onFail();
                }
                if (this.containers.error) {
                    this.containers.content.update(this.containers.error.innerHTML);
                }
            }
            return;
        }

        var d = r.response;
        var showing = (this.id == d.id || d.search);
        var buffer, metaData, view;

        if (showing) {
            buffer = this.buffer;
            metadata = this.metaData;
        } else {
            view = this.views[d.id];
            buffer = view.buffer;
            metadata = view.metaData;
            view.offset = opts.offset;
        }

        // Parse parameters
        buffer.update(d.msgs);
        metadata.set({
            cacheid: d.cacheid,
            label: d.label,
            sort_by: d.sortby,
            sort_dir: d.sortdir,
            sort_limit: d.sortlimit,
            special_folder: d.special,
            total_rows: d.totalrows,
            unseen: d.unseen
        });

        if (!showing) {
            return;
        }

        if (!this.scroller) {
            this.scroller = new ViewPort.Scroller(this);
            if (this.options.onFirstContent) {
                this.options.onFirstContent();
            }
        }

        if (!opts.options.noUpdateContent) {
            // If offset is null, we originally searched by ID. Thus, we need
            // to determine the offset by what was returned from the server.
            var offset = opts.offset;
            if (offset == null) {
                offset = buffer.getRows(buffer.getStartPosition(), 1);
                offset = offset.size() ? offset.first().rownum : 0;
            } else {
                // If we delete off the bottom of the page, our offset won't
                // account for this and we will have to do 2 updateContent()
                // calls - 1 now and 1 in scrollIdle().  Catch this case and
                // adjust the offset accordingly.
                offset = Math.min(offset, this.metaData.get('total_rows') - this.metaData.get('curr_page_size'));
            }
            this.updateContent(offset);
        }

        if (opts.onComplete) {
            opts.onComplete();
        }

        if (this.options.onComplete) {
            this.options.onComplete();
        }
    },

    updateContent: function(offset) {
        if (!this.buffer.isLoaded(offset)) {
            this.fetchBuffer(offset);
            return;
        }

        this.lastDisplayedStartPos = offset;
        this.lastDisplayedPageSize = this.metaData.get('page_size');
        var c = this.containers.content;

        this.clearMsgList();
        this.buffer.getRowsDiv(offset, this.lastDisplayedPageSize).each(function(r) {
            c.appendChild(r);
        });

        if (this.options.onContent) {
            this.options.onContent();
        }

        if (this.containers.empty && !c.immediateDescendants().size()) {
            c.update(this.containers.empty.innerHTML);
            return;
        }

        if (this.scroller) {
            this.scroller.updateSize();
            this.scroller.moveScroll(offset);
        }
    },

    checkCache: function() {
        DimpCore.doAction('CheckCache', { folder: this.id, cacheid: this.metaData.get('cacheid') }, [], this.checkCacheCallback.bind(this));
    },

    checkCacheCallback: function(r) {
        if (!r.response) {
            return;
        }
        var d = r.response;
        
        // TODO: PollFolders/CheckCache will automatically send the necessary
        //       message list information to save an AJAX call.
        // if (r.folder == this.folder && r.msglist) {
        //     this.viewport.ajaxUpdate({ response: r.msglist });
        // }
        if (d.folder && d.cacheexpire && d.folder == this.id) {
            this.reload();
        }

        if (this.options.onCheckCache) {
            this.options.onCheckCache(r);
        }
    },

    handleWait: function(call) {
        this.clearWait();

        // Server did not respond in defined amount of time.  Alert the
        // callback function and set the next timeout.
        if (call && this.options.onWait) {
            this.options.onWait();
        }

        // Call wait handler every 10 seconds
        this.waitHandler = setTimeout(function() { this.handleWait(true); }.bind(this), 10000);
    },

    clearWait: function() {
        if (this.waitHandler) {
            clearTimeout(this.waitHandler);
        }
    },

    clearMsgList: function() {
        var c = this.containers.content;
        // Don't use immediateDescendants() here since we want to clean out
        // anything inside of messagelist.
        $A(c.childNodes).each(function(n) {
            c.removeChild(n);
        });
    },

    clear: function() {
        this.clearMsgList();
        if (this.scroller) {
            this.scroller.clear();
        }
    },

    visibleRows: function() {
        return this.containers.content.immediateDescendants();
    },

    getRows: function(start, length) {
        return this.buffer.getRows(start, length);
    },

    getRowDiv: function(id) {
        var row = this.getRowById(id);
        return (row) ? $('msg' + row.uid) : row;
    },

    getMetaData: function(id) {
        return this.metaData.get(id);
    },

    setMetaData: function(vals) {
        return this.metaData.set(vals);
    },

    currentOffset: function() {
        return (this.scroller) ? this.scroller.currentOffset() : 0;
    },

    updateFlag: function(id, flag, add, permanent) {
        var row = this.getRowById(id);
        if (!row) {
            return;
        }

        if (permanent) {
            row.bg = (add) ? $w(row.bg).concat([ flag ]).join(' ') : $w(row.bg).without([ flag ]).join(' ');
        }

        row = this.getRowDiv(id);
        if (row) {
            if (add) {
                row.addClassName(flag);
            } else {
                row.removeClassName(flag);
            }
        }
    },

    bufferSize: function() {
        return this.buffer.size;
    }

};

/**
 * ViewPort.MetaData
 */
ViewPort.MetaData = Class.create();
ViewPort.MetaData.prototype = {
    // Vars used: see get() and set()
    largeBufferSize: 7.0, // 7 pages
    nearLimitFactor: 0.2,  // 20% of buffer

    initialize: function(pageSize, totalRows) {
        this.set({ page_size: pageSize, total_rows: totalRows });
    },

    get: function(id) {
        switch (id) {
        case 'cacheid':
        case 'label':
        case 'largeBufferSize':
        case 'nearLimitFactor':
        case 'page_size':
        case 'sort_by':
        case 'sort_dir':
        case 'sort_limit':
        case 'total_rows':
            return this[id];

        case 'special_folder':
        case 'unseen':
            return (this[id]) ? this[id] : 0;

        case 'limit_tolerance':
            return Math.round(this.get('buffer_size') * this.nearLimitFactor);
        
        case 'buffer_size':
            return Math.round(this.largeBufferSize * this.page_size);

        case 'curr_page_size':
            return Math.min(this.page_size, this.total_rows);

        default:
            return null;
        }
    },

    set: function(vals) {
        $H(vals).each(function(node) {
            switch (node.key) {
            case 'buffer_size':
            case 'page_size':
            case 'sort_by':
            case 'sort_dir':
            case 'sort_limit':
            case 'special_folder':
            case 'total_rows':
            case 'unseen':
                this[node.key] = parseInt(node.value);
                return;

            case 'largeBufferSize':
            case 'nearLimitFactor':
                this[node.key] = parseFloat(node.value);
                return;

            case 'cacheid':
            case 'label':
                this[node.key] = node.value;
                return;
            }
        }.bind(this));
    }

};

/**
 * ViewPort.Scroller
 */
ViewPort.Scroller = Class.create();
ViewPort.Scroller.prototype = {
    // Vars used and defaulting to null:
    //   scrollTimeout

    initialize: function(parent) {
        this.parent = parent;
        this.lineHeight = parent.options.lineHeight ? parent.options.lineHeight : 20;
        this.createScrollBar();
    },

    createScrollBar: function() {
        var c = this.parent.containers.content;

        // Create the outer div.
        this.scrollerDiv = $(Builder.node('DIV', { className: 'sbouterdiv', id: 'sbouterdiv', style: 'height:' + c.getHeight() + 'px;' }));
        Event.observe(this.scrollerDiv, 'scroll', this.handleScroll.bindAsEventListener(this));

        // Give our parent a right margin just big enough to
        // accomodate the scrollbar.
        c.setStyle({ marginRight: '-19px' });

        // Create the inner div.
        this.heightDiv = $(Builder.node('DIV', { className: 'sbinnerdiv', id: 'sbinnerdiv' }));
        this.updateSize();

        this.scrollerDiv.appendChild(this.heightDiv);
        c.parentNode.insertBefore(this.scrollerDiv, c.nextSibling);

        // Mouse wheel handler.
        Event.observe(c, Prototype.Browser.IE ? 'mousewheel' : 'DOMMouseScroll', function(e) {
            this.scrollerDiv.scrollTop += (e.wheelDelta >= 0 || e.detail < 0)
                // wheel-up.
                ? (-2 * this.lineHeight)
                : (2 * this.lineHeight);
            this.handleScroll();
            Event.stop(e);
        }.bindAsEventListener(this), false);
    },

    onResize: function() {
        var h = this.parent.containers.content.getHeight();
        var newPageSize = Math.floor(h / this.lineHeight);
        if (newPageSize == this.parent.metaData.get('page_size')) {
            return;
        }
        this.parent.metaData.set({ page_size: newPageSize });

        // Update the outer div.
        this.scrollerDiv.style.height = h + 'px';

        // Update the inner div.
        this.updateSize();

        // Update displayed content.
        this.parent.requestContentRefresh(this.currentOffset());
        this.handleScroll();
    },

    updateSize: function() {
        this.heightDiv.style.height = Math.round(this.parent.containers.content.getHeight() * this.parent.metaData.get('total_rows') / this.parent.metaData.get('page_size')) + 'px';
    },

    clear: function() {
        var h = this.parent.containers.content.getHeight();
        var newPageSize = Math.floor(h / this.lineHeight);
        this.parent.metaData.set({ page_size: newPageSize });
        this.scrollerDiv.style.height = this.heightDiv.style.height = h + 'px';
    },

    moveScroll: function(offset) {
        this.scrollerDiv.scrollTop = Math.round((offset / this.parent.metaData.get('total_rows')) * this.heightDiv.getHeight());
        this.handleScroll();
    },

    handleScroll: function(e) {
        if (this.scrollTimeout) {
            clearTimeout(this.scrollTimeout);
        }

        if (this.parent.options.onScroll) {
            this.parent.options.onScroll();
        }

        if (e) {
            this.scrollTimeout = setTimeout(this.scrollIdle.bind(this), 100);
        } else {
            this.scrollIdle();
        }
    },

    scrollIdle: function() {
        if (this.scrollTimeout) {
            clearTimeout(this.scrollTimeout);
        }

        this.parent.requestContentRefresh(this.currentOffset());

        if (this.parent.options.onScrollIdle) {
            this.parent.options.onScrollIdle();
        }
    },

    currentOffset: function() {
        var offset = Math.min(this.parent.metaData.get('total_rows'),
                              Math.round(this.scrollerDiv.scrollTop *
                                         this.parent.metaData.get('total_rows') / this.heightDiv.getHeight()));
        return isNaN(offset) ? 0 : offset;
    }

};

/**
 * ViewPort.Buffer
 */
ViewPort.Buffer = Class.create();
ViewPort.Buffer.prototype = {
    // Vars used and defaulting to null:
    //   updated
    startPos: 0,
    size: 0,
    rows: [],

    initialize: function(parent) {
        this.parent = parent;
        this.template = new Template(parent.options.template);
    },

    update: function(r) {
        this.updated = true;
        this.size = this.startPos = 0;

        if (!r) {
            this.rows.clear();
            return;
        }

        this.rows = r;
        this.size = this.rows.size();
        this.startPos = this.size ? parseInt(this.rows.first().rownum) : 0;
    },

    isLoaded: function(offset) {
        return (this.updated &&
                (offset >= this.startPos) &&
                Math.min(offset + this.parent.metaData.get('page_size'), this.parent.metaData.get('total_rows')) <= (this.startPos + this.size));
    },

    isNearingTopLimit: function(offset) {
        return offset - this.startPos < this.parent.metaData.get('limit_tolerance');
    },

    isNearingBottomLimit: function(offset) {
        return (this.startPos + this.size - offset - this.parent.metaData.get('page_size')) < this.parent.metaData.get('limit_tolerance');
    },

    isAtTop: function() {
        return this.startPos === 0;
    },

    isAtBottom: function() {
        return (this.startPos + this.size) == this.parent.metaData.get('total_rows');
    },

    isNearingLimit: function(offset) {
        if (this.size == this.parent.metaData.get('total_rows')) {
            return false;
        }
        return (!this.isAtTop() && this.isNearingTopLimit(offset)) ||
               (!this.isAtBottom() && this.isNearingBottomLimit(offset));
    },

    getRowById: function(id) {
        return this.rows.find(function(r) {
            return (r.uid == id);
        });
    },

    getRows: function(start, length) {
        var s = start - this.startPos;
        if (s < 0 || s >= this.size) {
            return [];
        }

        var e = Math.min(s + length, this.size);
        if (e < s) {
            return [];
        }

        return this.rows.slice(s, e);
    },

    getRowsDiv: function(start, length) {
        var div = $(Builder.node('DIV'));
        this.getRows(start, length).each(function(r) {
            new Insertion.Bottom(div, this.template.evaluate(r));
        }.bind(this));

        return div.immediateDescendants();
    },

    getStartPosition: function() {
        return this.startPos;
    }

};

/**
 * ViewPort.Filter
 */
ViewPort.Filter = Class.create();
ViewPort.Filter.prototype = {
    // Vars used and defaulting to null/false:
    //   filtering, lastFilter, observer, uf
    uf: {},

    initialize: function(parent, input, action, callback) {
        this.parent = parent;
        this.input = input;
        this.action = action;
        this.callback = callback || false;

        if (this.input) {
            Event.observe(this.input, 'keyup', this.onKeyUp.bind(this));
        }
    },

    onKeyUp: function() {
        if (this.observer) {
            clearTimeout(this.observer);
        }
        this.observer = setTimeout(this.filter.bind(this), 500);
    },

    filter: function(force) {
        var val = $F(this.input).toLowerCase();
        if (!val) {
            if (this.filtering) {
                this.parent.action = this.uf.action;
                this.parent.buffer = this.uf.buffer;
                this.parent.lastDisplayedStartPos = -1;
                this.parent.metaData = this.uf.metadata;
                if (this.parent.requestContentRefresh(this.uf.offset)) {
                    this.checkCache();
                    if (this.parent.options.onComplete) {
                        this.parent.options.onComplete();
                    }
                }
              
                this.filtering = this.lastFilter = null;
            }
            return;
        }

        if (!force && (val == this.lastFilter)) {
            return;
        }
        this.lastFilter = val;

        if (!this.filtering) {
            this.filtering = true;

            // Store old parameters.
            this.uf.action = this.parent.action;
            this.uf.offset = this.parent.scroller.currentOffset();
            this.uf.buffer = Object.clone(this.parent.buffer);
            this.uf.metadata = Object.clone(this.parent.metaData);

            // Set the search URL.
            this.parent.action = this.action;

            // Filter visible rows immediately.
            var c = this.parent.containers.content;
            c.immediateDescendants().each(function(n) {
                if (n.getText(true).toLowerCase().indexOf(val) == -1) {
                    n.remove();
                }
            });
            if (this.parent.containers.empty && !c.immediateDescendants().size()) {
                c.update(this.parent.containers.empty.innerHTML);
            }
        }

        // Get parameters from a callback function, if defined.
        var c_params = (this.callback) ? $H(this.callback()) : new Hash();
        if (this.uf.params) {
            c_params = c_params.merge(this.uf.params);
        }

        // Set search.
        c_params.filter = val;

        this.parent.fetchBuffer(0, c_params);
    }

};
